const DEBUG_NAME = "Thumbs singleton";

function myDump (msg)
{
  var fullmsg = DEBUG_NAME + ": " + msg;

  dump(fullmsg + "\n");

  try {
    var cns = Components.
                classes["@mozilla.org/consoleservice;1"].
                  getService(Components.interfaces.nsIConsoleService);
    cns.logStringMessage(fullmsg);
  }
  catch (e) {
    dump("Can't put that in the JS console.\n");
  }
}


/******************************************************************************/


/******************************************************************************/

function MyComponent() {}

MyComponent.prototype = {

    QueryInterface: function (iid) {
      if (!(
		        iid.equals(Components.interfaces.nsISupports)
		     || iid.equals(Components.interfaces.nsIThumbsSingleton)
		      ))
          throw Components.results.NS_ERROR_NO_INTERFACE;

      return this;
    },
    

    tgpWindow : null,

    getTGPWindow : function() {
      var temp = this.tgpWindow;
      
      // keeping this reference around would be a (limited?) leak. 
      // in previous versions, this line was commneted out due to TBE issues.
      // temporarily commented out to test something out.
      this.tgpWindow = null; 
        
      return temp;
    },
    
    setTGPWindow : function(w) {
      // assert(this.tgpWindow == null) ?
      this.tgpWindow = w;
    }
}

/******************************************************************************/

var myModule = {

    myClassID:     Components.ID("{2b537acc-09d7-4b4c-bf88-836ee33d24c3}"),
    myContractID:  "@squarefree.com/thumbs-singleton;1",
    myClassName:   "thumbs-singleton",

    firstTime  : true,

    registerSelf: function (compMgr, fileSpec, location, type) {
      //myDump(" ***** Registering: " + DEBUG_NAME + " ****");

      compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
      compMgr.registerFactoryLocation(this.myClassID,
                                      this.myClassName,
                                      this.myContractID,
                                      fileSpec,
                                      location,
                                      type);
    },

  	unregisterSelf: function(compMgr, fileSpec, location) {
      //dump(" ***** Unregistering: " + DEBUG_NAME + " ****\n"); // In shutdown mode, use normal dump.

      compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  		compMgr.unregisterFactoryLocation(this.myClassID, fileSpec);
    },

    getClassObject: function (compMgr, cid, iid) {

        if (!cid.equals(this.myClassID))
            throw Components.results.NS_ERROR_NO_INTERFACE;

        if (!iid.equals(Components.interfaces.nsIFactory))
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

        return this.myFactory;
    },

    myFactory: {
        mySingleton: null,
        
        getService : function(outer, iid) {
            //myDump("getService called with interface id: " + iid);
            if (outer != null)
                throw Components.results.NS_ERROR_NO_AGGREGATION;

            if (!this.mySingleton)
            {
              //myDump("CREATING A ***NEW*** INSTANCE");
              this.mySingleton = new MyComponent();
            }

            return this.mySingleton.QueryInterface(iid);
        },

        createInstance: function (outer, iid) {
          return this.getService(outer, iid);
        }
    },

    canUnload: function(compMgr) {
        //dump("****** About to unload: " + DEBUG_NAME + " *****\n"); // In shutdown mode, use normal dump.
        
        // "It is very important to ensure that no outstanding references to 
        // the module's code/data exist before returning true."
        this.mySingleton = null;
        
        return true;
    }
};

function NSGetModule(compMgr, fileSpec) { return myModule; }

/******************************************************************************/