/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is thumbs for Firefox.
 *
 * The Initial Developer of the Original Code is Jesse Ruderman.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/***********************************************************************************************/

/*
 * Globals.
 */

var referrerFixer;

var tgpURL;

var unvisitedGalleries; // array of structures: { uri : undefined, infoBlock : undefined }
var visitedGalleries;

// http://docs.mandragor.org/files/Misc/Mozilla_applications_en/mozilla-chp-11-sect-3.html
var bundle;

/***********************************************************************************************/

/*
 * Cached services (which are also globals).
 */

var ioService = Components.
			            classes["@mozilla.org/network/io-service;1"].
			              getService(Components.interfaces.nsIIOService);

// http://lxr.mozilla.org/aviarybranch/source/toolkit/components/history/public/nsIBrowserHistory.idl
// inherits:
// http://lxr.mozilla.org/aviarybranch/source/docshell/base/nsIGlobalHistory2.idl
var browserHistory = Components.
	                     classes["@mozilla.org/browser/global-history;2"].
	                       getService(Components.interfaces.nsIBrowserHistory);

// Detect "old" (Firefox 1.0) version of nsIGlobalHistory2,
var browserHistoryOldVersion = Components.interfacesByID["{c4d143b7-22d7-4760-b9a0-ac942016f189}"]; 

/***********************************************************************************************/

/*
 * Onload statements.
 */

window.onload = function() {

  bundle = srGetStrBundle("chrome://thumbs/locale/thumbs.properties");

  // Warn the user if images are disabled.
  // It would be better if we could override this pref, or if it didn't apply to chrome, or something.
  // Is this the cost of living in the content area?
  if (getIntPref("network.image.imageBehavior", 0) == 2)
    alert(bundle.GetStringFromName("ImagesDisabled"));

  var tgpWindow = Components.
                    classes["@squarefree.com/thumbs-singleton;1"].
                      getService(Components.interfaces.nsIThumbsSingleton).
                        getTGPWindow();
            
  tgpURL = tgpWindow.location.href; // safe because gecko special-cases location.href to be un-screw-with-able
  
  document.title = bundle.formatStringFromName("ThumbsFrom", [tgpURL], 1);

  var addresses = getGalleryAddresses(tgpWindow);
 
  var message = document.createTextNode(addresses.message);
  var para = document.createElement("p");
  para.appendChild(message);
  document.body.appendChild(para);

  splitVisited(addresses);

  buildLinkBlocks(unvisitedGalleries);
  
  var numUnvisited = unvisitedGalleries.length;
  var numVisited = visitedGalleries.length;
  
  createButton(bundle.formatStringFromName("MarkAllVisited", [numUnvisited], 1), markAllVisited, numUnvisited);
  document.body.appendChild(document.createElement("hr"));
  createButton(bundle.formatStringFromName("ShowVisited", [numVisited], 1), showVisited, numVisited);
  
/*
  para = document.createElement("p");
  document.body.appendChild(para);
  para.innerHTML = 'Thumbs is free software written by <a href="http://www.squarefree.com/">Jesse Ruderman</a>, who appreciates <a href="mailto:jruderman@gmail.com">feedback</a> and <a href="http://www.squarefree.com/extensions/thumbs-a/donate">donations</a>.';
*/
}


/***********************************************************************************************/

/*
 * Helper functions.
 */

function markAllVisited()
{
  var i;

  // Add the addresses to global history.
  for (i = 0; i < unvisitedGalleries.length; ++i) {
    var address = unvisitedGalleries[i].uri;
    // The following line can fail if history.dat is read-only.  Oh well.
    if (!browserHistory.isVisited(address)) {
      if (browserHistoryOldVersion)
        browserHistory.addURI(address, false, false);
      else
        browserHistory.addURI(address, false, false, null);
    }
  }

  // Hack to force Firefox to update link colors.
  // This shouldn't be necessary - http://bugzilla.mozilla.org/show_bug.cgi?id=78510.
  for (i = 0; i < document.links.length; ++i) {
    var link = document.links[i];
		var temp = link.getAttribute("href");
		link.href = "";
		link.href = temp;
  }  
}

/*
 * Initializes visitedGalleries and unvisitedGalleries from allAddresses (as a side effect).
 */
function splitVisited(allAddresses)
{
  visitedGalleries = [];
  unvisitedGalleries = [];

  function isVisited(address) {
    try {
      return browserHistory.isVisited(address);
    }
    catch(e) {
      // When history.dat is read-only, isVisited throws NS_ERROR_NOT_INITIALIZED.
      // Silently (!) treat this as an unvisited link.
      return false;
    }
  }

  for (var i = 0; i < allAddresses.length; ++i) {
    var address = allAddresses[i];
    if (isVisited(address))
      visitedGalleries.push({uri: address});
    else
      unvisitedGalleries.push({uri: address});
  }
}

function showVisited()
{
  buildLinkBlocks(visitedGalleries);
}

function createButton(buttonText, onclick, enabled)
{
  var para = document.createElement("p");
  var button = document.createElement("button");
  var textNode = document.createTextNode(buttonText);

  button.onclick = onclick;
  button.disabled = !enabled;
  
  button.appendChild(textNode);
  para.appendChild(button);
  document.body.appendChild(para);
}

/*
 * stringToURI: converts URL (string) to URI (nsIURI object), relative to base URI (if provided).
 * Returns null for bogus or non-http URLs!  Caller must check!
 */
function stringToURI(url, base)
{
  var uri;

  try {
	  uri = ioService.newURI(url, null, base || null);
	}
	catch(er) { 
	  // aim: URLs, totally bogus URLs resulting from "remove redirects" on a javascript: link, etc.
	  return null;
	}

  if (uri.schemeIs("http")) // for security (e.g. javascript: and data: URLs) and sanity (e.g. XMLHttpRequest)
    return uri;
  
  return null;
}

function removeDuplicates(addresses)
{
  var seen = []; // Hash with keys of the form (" " + url) and values from {true, undefined}.
  var ret = []; // Subset of addresses
  
  for (var i = 0; i < addresses.length; ++i) {
    var uri = addresses[i];
    var url = uri.spec;
  
    if (!seen[" " + url]) {
      ret.push(uri);
      seen[" " + url] = true;
    }
  }
  
  return ret;
}

function clobberInfoBlock(infoBlock)
{
  // Instead of removing previous children, hide them, so View Selection Source can recover them!
  if (infoBlock.lastChild)
    infoBlock.lastChild.style.display = "none";
}

function displayMessage(infoBlock, className, text)
{
  clobberInfoBlock(infoBlock);

  var span = document.createElement("span");
  span.appendChild(document.createTextNode(text));
  span.className = className; 
  
  infoBlock.appendChild(span);
}


function loadImage(galleryURI, infoBlock, srcURI) // combine first two params into gallery?
{
  clobberInfoBlock(infoBlock);
  
  referrerFixer.srcToReferrer[" " + srcURI.spec] = galleryURI;
  
  var img = document.createElement("img");
  img.style.maxHeight = getIntPref("extensions.thumbs.gallerylink.thumbnailHeight", 230) + "px";
  img.style.maxWidth = getIntPref("extensions.thumbs.gallerylink.thumbnailWidth", 400) + "px";

  img.src = srcURI.spec;
  
  img.alt = bundle.GetStringFromName("BrokenImage"); // Lame.  Isn't there some way to make a broken image icon appear?

  var span = document.createElement("span");
  span.appendChild(img);
  span.className = "error"; // for alt text
  
  infoBlock.appendChild(span);
}

// Remove Redirects bookmarklet, with the "see if it has a mouseover handler" parts taken out due to security paranoia.
// Want the "see if it has a mouseover handler" behavior?  Use the bookmarklet, then use the extension :)
function removeRedirects(url)
{
  var t,i,j;

  t=url.replace(/[%]3A/ig,':').replace(/[%]2f/ig,'/');
  i=t.lastIndexOf('http://');
  if(i>0){
    t=t.substring(i);
    j=t.indexOf('&');
    if(j>0)
      t=t.substring(0,j);
    return unescape(t)
  }

  return url;
}


/***********************************************************************************************/


/*
 * Step 1: Extract URLs.  This has to be done carefully, because we're dealing with an untrusted DOM.  See
 * http://www.squarefree.com/securitytips/mozilla-developers.html#untrusted-dom for more info.
 * Also strips out any duplicate links.
 */
function getGalleryAddresses(win)
{
  // Non-selected addresses, selected addresses
  var addresses = [[], []];

  var selection = nativeMethod(win, "getSelection")();
  var collapsed = nativeGet(selection, "isCollapsed");
  var selectionContainsNode = nativeMethod(selection, "containsNode");
  
	var doc = nativeGet(win, "document");
	var links = nativeGet(doc, "links");

	for (var i = 0; i < links.length; ++i) {
    // No "native" wrapper on the following line!
		// Somehow, |links| is an ordinary, not-screwed-with HTMLCollection.
		// This is good, because I can't use nativeGet with numbers.
		// Still, this line feels sketchy and could be a security hole.
	  var link = links[i];
	  
    var url = nativeGet(link, "href"); // this better be a string
  
    // Tiny hard-coded blacklist:
    //   Clickzs is 50% of ad links on some sites. 
    //   Google "cache" is useless and gives the wrong referrer when you click links from it.
    //   Google "similar pages" is useless.
	  if (!url.match(/clickzs|cache\:|related\:/)) {
		  var uri = stringToURI(removeRedirects(url));
	
		  if (uri)
        addresses[(!collapsed && selectionContainsNode(link, true)) ? 1 : 0].push(uri);
		}
  } 
  
  if (addresses[1].length) {
    // Some links were selected.  Return selected links.
    addresses[1] = removeDuplicates(addresses[1]);
    addresses[1].message = bundle.formatStringFromName("SelectedLinks", [addresses[1].length, tgpURL], 2);
    return addresses[1];
  }
  else {
    // No links were selected.  Return all links.
    addresses[0] = removeDuplicates(addresses[0]);
    addresses[0].message = bundle.formatStringFromName("NoSelectedLinks", [addresses[0].length, tgpURL], 2);
    return addresses[0];
  }
}



function cleanUp()
{
  referrerFixer.finishObserving();
  
  // paranoid leak prevention: clear all globals
  // (i don't really know what i'm doing; this might not be necessary.)
  referrerFixer = null;
  tgpURL = null;
  unvisitedGalleries = null;
  visitedGalleries = null;
  bundle = null;
  ioService = null;
  browserHistory = null;
  browserHistoryOldVersion = null;
}


/*
 * Step 2: Build the page.
 */
function buildLinkBlocks(galleries)
{
  referrerFixer = new ReferrerFixer();
  window.onbeforeunload = cleanUp;

  for (var i = 0; i < galleries.length; ++i) {
  
    var uri = galleries[i].uri;

    var link = document.createElement("a");
    
    link.className = "gallink";
    
    var height = 
      getIntPref("extensions.thumbs.gallerylink.thumbnailHeight", 230) +
      getIntPref("extensions.thumbs.gallerylink.extraHeight", 20);
    
    link.style.height = height + "px" ;

    link.href = uri.spec;
    
    link.onclick = linkClick;

    // The link consists of two inner blocks: an info block, showing a thumbnail or status message,
    // and a hostname block, showing the hostname of the link.
    
    var infoBlock = document.createElement("div");
    link.appendChild(infoBlock);
    displayMessage(infoBlock, "informational", bundle.GetStringFromName("Pending"));

    var hostBlock = document.createElement("div");
    link.appendChild(hostBlock);
    var linkText = document.createTextNode(uri.host);
    hostBlock.appendChild(linkText);
    
    document.body.appendChild(link);
    
    galleries[i].infoBlock = infoBlock;
  }

  var numImmediate = getIntPref("extensions.thumbs.numImmediate", 10);

  // Kick off numImmediate loads immediately.
  for (var i = 0; i < numImmediate && i < galleries.length; ++i)
    requestGallery(galleries, i);

  // Do the rest of the loads on timers.
  grabAnother(galleries, numImmediate);
}

function grabAnother(galleries, i)
{
  if (i < galleries.length) {

    requestGallery(galleries, i);

    setTimeout(grabAnother,
               getIntPref("extensions.thumbs.delay.milliseconds", 90),
               galleries, i + 1);
               // We're called one extra time, oh well. We'd need the |if| anyway.
  }
}

function requestGallery(galleries, i)
{
  var uri = galleries[i].uri;
  var infoBlock = galleries[i].infoBlock;

  try{
    displayMessage(infoBlock, "informational", bundle.GetStringFromName("Loading"))

	  var request = new XMLHttpRequest();

	  request.onload = function() {
	    parseGallery(uri, request, infoBlock); 
	  };

	  request.onerror = function(e) {
	    // This is called for connection errors like "host not found", but parseGallery gets server errors like 404.
	    displayMessage(infoBlock, "error", bundle.GetStringFromName("HTTPConnectionError"));
	  };

	  request.open("GET", uri.spec);
	  request.send(null);
	}
	catch(e) { 
	  displayMessage(infoBlock, "error", bundle.formatStringFromName("XMLHttpRequestError", [uri.spec, e], 2));
	}
}

function parseGallery(galleryURI, request, infoBlock)
{
  if (request.status != 200) {
    // If it's not 200 (OK), assume it's an error.
    // This can lead to silliness like "HTTP error 302" if a site uses a malformed (?) redirect.
    // And a site could probably get us to display "HTTP error 666" - a major security hole :)
    displayMessage(infoBlock, "error", bundle.formatStringFromName("HTTPError", [request.status], 1));
    return;
  }

  function parseAttribute(link, regexp)
  {
    try{
	    if (!link.match(regexp))
	      return null;
	
	    var attrplus = RegExp.$2	; 
	    var quoteChar = attrplus[0]; 
	    var attrlen, attr;
	    if (quoteChar == '"' || quoteChar == "'") { 
	      attrlen = attrplus.slice(1).indexOf(quoteChar); 
	      attr = attrplus.substr(1, attrlen);
	    }
	    else {
	      attrlen = Math.min(attrplus.indexOf(" "), attrplus.indexOf(">"));
	      attr = attrplus.substr(0, attrlen);
	    }
	    
	    return attr;
	  }
	  catch(er) { 
	    // Probably due to a combination of my bad arithmetic and the page's malformed HTML.  Fail happily.
	    return null;
	  }
  }
  
  // We want to choose an image that's a thumbnail.
  function thumbiness(url)
  {
    if(url.match(/\.(jpg|gif|png|jpeg|rm|mov|wmv|avi|mpg|mpeg)$/i))
      return 2;
    
    if(url.match(/\.(htm|html|php)$/i))
      return 1;
      
    return 0;
  }

  // We want to choose an image that links to something in the same directory.
  function longestCommonPrefixLength(a, b)
	{
	  for (var i = 0; i < a.length && i < b.length && a[i] == b[i]; ++i)
	    ;
	
	  return i;
	}

  var html = request.responseText;

  /*
   * I'm lame and use regexps.
   *
   * I use (.|\r|\n) instead of . because . doesn't match \n (why??) or \r (not even documented!).
   *
   * I use *? instead of * to avoid eating too much.
   */

  var links = html.match(/\<a\s(.|\r|\n)*?\<\/a\>/ig); 
  var i,link;

  var bestT = -1;
  var bestP = -6;
  var bestSrc;
  
  var galleryURL = galleryURI.spec;
  
  if (links) {
    //displayMessage(infoBlock, "informational", links.length + " links");
	  for (i=0; link=links[i]; ++i) { 

      var src, href;

	    if ( (src = parseAttribute(link, /\<img\s(.|\r|\n)*?src\=((.|\r|\n)*)$/i))
	      && (href = parseAttribute(link, /\<a\s(.|\r|\n)*?href\=((.|\r|\n)*)$/i)) ) {
	      
        var t = thumbiness(href); /* ok to send a relative link to thumbiness() */
        
        var hrefURI = stringToURI(href, galleryURI);
        if (hrefURI) {
          var p = longestCommonPrefixLength(hrefURI.spec, galleryURL);
	        if (t > bestT || (t == bestT && p > bestP + 3)) {  
	          bestT = t;
	          bestP = p
		        bestSrc = src;
		      }
		    }
      }
	  }
	}

  if (!bestSrc) {
    displayMessage(infoBlock, "error", bundle.GetStringFromName("NoImageLinks"))
    return;
  }

  var srcURI = stringToURI(bestSrc, galleryURI);

  if (!srcURI) {
    displayMessage(infoBlock, "error", bundle.GetStringFromName("ImageBogusURI"));
    return;
  }

  loadImage(galleryURI, infoBlock, srcURI);
}


function openInTab(url)
{
  var ww = Components
             .classes["@mozilla.org/embedcomp/window-watcher;1"]
             .getService(Components.interfaces.nsIWindowWatcher);
             
  ww.activeWindow.getBrowser().addTab(url);
}

function linkClick(ev)
{
  // If a normal left-click or alt-click...
  if (ev.button == 0 && !ev.shiftKey && !ev.ctrlKey && !ev.metaKey) { 

    // Open in tab or window depending on pref
    if (getBoolPref("extensions.thumbs.opentabfor.gallerylink", true)) {
      openInTab(this.href);
    }
    else {
      window.open(this.href);
    }

    // don't do anything else
    ev.preventDefault();
    ev.stopPropagation();

  }
}